/*
    Font filters implementation.
    Copyright (C) 2020 2021  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */


/***************************************************************************/
/* include vari */

#include <stdio.h>			// fprintf()
#include <stdlib.h>			// malloc()
#include <string.h>			// memcpy()
#include "fontdesc.h"			// font data type

/***************************************************************************/
/* define interni alla libreria */

/***************************************************************************/
/* strutture contenute all'interno della libreria */

/***************************************************************************/
/* costanti interne alla libreria */

/***************************************************************************/
/* costanti pubbliche */
			
/***************************************************************************/
/* variabili interne alla libreria */

/***************************************************************************/
/* funzioni interne alla libreria */

static uint8_t ByteCollapse(uint8_t Data,uint32_t CollPt);
  /* Collapses byte shape deleting bit CollPt. */

static uint8_t ByteFlip(uint8_t Data);
  /* Inverte l'ordine dei bit all'interno del byte. */

static void Cw(uint8_t *In,uint8_t *Out);
  /* Ruota verso destra la shape puntata da In e la inserisce in Out. */

static void Ccw(uint8_t *In,uint8_t *Out);
  /* Ruota verso sinistra la shape puntata da In e la inserisce in Out. */

/***************************************************************************/
/* funzioni */

static uint8_t ByteCollapse(uint8_t Data,uint32_t CollPt)
{
  /* Collapses byte shape deleting bit CollPt. */
  
  
  
  uint32_t	Bit=0;
  uint8_t	HoldMask=0;
  uint8_t	Collapsed;
  
  
  
  // la hold mask deve avere tutti 1 a destra fino al bit CollPt
  while(Bit < CollPt)
  {
    HoldMask|=1 << Bit;
    Bit++;
  }
  
  // elimina il bit CollPt-esimo e collassa il resto verso destra
  Collapsed =Data & HoldMask;			// trattiene i bit fermi a destra
  Collapsed|=(Data >> 1) & (~HoldMask);		// aggiunge la parte collassata
  return(Collapsed);				// termina
}



static uint8_t ByteFlip(uint8_t Data)
{
  /* Inverte l'ordine dei bit all'interno del byte. */
  
  
  
  unsigned int	i=8;
  uint8_t	Out;
  
  
  
  do
  {
    Out>>=1;					// fa posto al nuovo bit
    Out|=Data & 0x80;				// sposta l'MSB Data -> Out
    Data<<=1;					// shifta Data
  }while(--i);
  return(Out);					// termina
}



static void Cw(uint8_t *In,uint8_t *Out)
{
  /* Ruota verso destra la shape puntata da In e la inserisce in Out. */
  
  
  
  unsigned int	i,j;
  uint8_t	Data;
  
  
  
  for(j=0;j<8;j++)				// indice bit
  {
    for(i=0;i<8;i++)				// indice byte
    {
      Data<<=1;					// fa posto da LSB
      Data|=((In[i]<<j) & 0x80)>>7;		// aggiunge il j-esimo bit da In
    }
    *Out++=Data;				// aggiorna il nuovo font
  }
}



static void Ccw(uint8_t *In,uint8_t *Out)
{
  /* Ruota verso sinistra la shape puntata da In e la inserisce in Out. */
  
  
  
  unsigned int	i,j;
  uint8_t	Data;
  
  
  
  for(j=0;j<8;j++)				// indice bit
  {
    for(i=0;i<8;i++)				// indice byte
    {
      Data>>=1;					// fa posto da MSB
      Data|=((In[i]>>j) & 0x01)<<7;		// aggiunge il j-esimo bit da In
    }
    *Out++=Data;				// aggiorna il nuovo font
  }
}





FontDesc_t *NewFont(const FontDesc_t *Font)
{
  /* Allocates a new modifiable font. */
  
  
  
  FontDesc_t	*NewFont;
  uint8_t	*Data;
  
  
  
  if(Font == NULL)				// se non c'e'
    return(NULL);				// propaga NULL
  
  // alloca il nuovo font in RAM
  if((NewFont=malloc(sizeof(FontDesc_t) + FontDataSize(Font))) == NULL)
    return(NewFont);				// termina con fallimento
  
  // copia il font originale in RAM
  memcpy(NewFont,Font,sizeof(FontDesc_t));	// copia il descriptor
  Data=(uint8_t*)&NewFont[1];			// punta oltre il descriptor
  NewFont->Data=Data;				// modifica il puntatore
  memcpy(Data,Font->Data,FontDataSize(Font));	// copia il font
  return(NewFont);				// termina con successo
}



FontDesc_t *RightShift(FontDesc_t *Font,uint32_t Shift)
{
  /* Right shift all font data. */
  
  
  
  uint8_t	*Data=Font->Data;		// punta al data array
  uint32_t	Len=FontDataSize(Font);		// dimensione data array
  
  
  
  if(Font == NULL)				// se non c'e'
    return(NULL);				// propaga NULL
  do
  {
    *Data++>>=Shift;				// data shift
  }while(--Len);				// fino a fine array
  return(Font);					// termina
}



FontDesc_t *LeftShift(FontDesc_t *Font,uint32_t Shift)
{
  /* Left shift all font data. */



  uint8_t	*Data=Font->Data;		// punta al data array
  uint32_t	Len=FontDataSize(Font);		// dimensione data array
  
  
  
  do
  {
    *Data++<<=Shift;				// data shift
  }while(--Len);				// fino a fine array
  return(Font);					// termina
}



FontDesc_t *FlipLeftToRight(FontDesc_t *Font)
{
  /* Flips characters left to right. */



  uint8_t	*Data=Font->Data;		// punta al data array
  uint32_t	Len=FontDataSize(Font);		// dimensione data array
  
  
  
  if(Font == NULL)				// se non c'e'
    return(NULL);				// propaga NULL
  do
  {
    *Data=ByteFlip(*Data);			// flip left to right
    Data++;					// avanti Savoia!!
  }while(--Len);				// fino a fine array
  return(Font);					// termina
}
  


FontDesc_t *RotateRight(FontDesc_t *Font)
{
  /* Rotates all characters shapes 90 degrees clock-wise. */



  FontDesc_t	*Copy;
  uint8_t	*InData=Font->Data;		// punta al data array
  uint8_t	*OutData;
  uint8_t	A;
  uint32_t	Len=FontNumOfChars(Font);	// # di char nel font
  
  
  
  if(Font == NULL)				// se non c'e'
    return(NULL);				// propaga NULL
  if(Font->BytePerChar != 8)			// se non ha 8 byte per char
    return(Font);				// non possiamo fare niente
  
  // crea una nuova copia del font per poi distruggere l'originale
  if((Copy=NewFont(Font)) == NULL)		// se non riesce a copiare
  {
    fprintf(stderr,"%s: unable to copy original font.\n",__FUNCTION__);
    free(Font);					// distrugge comunque l'origin.
    return(NULL);				// termian con fallimento
  }

  // scambia width ed height
  A=		Copy->Width;
  Copy->Width=	Copy->Height;
  Copy->Height=	A;
  
  // modifica tutti i char uno alla volta
  OutData=Copy->Data;				// punta alla destinazione
  do
  {
    Cw(InData,OutData);				// ruota a destra
    InData+=Font->BytePerChar;			// prossimo char
    OutData+=Font->BytePerChar;
  }while(--Len);				// fino a fine array
  free(Font);					// distrugge l'originle
  return(Copy);					// termina
}



FontDesc_t *RotateLeft(FontDesc_t *Font)
{
  /* Rotates all characters shapes 90 degrees counter-clock-wise. */
  
  
  
  FontDesc_t	*Copy;
  uint8_t	*InData=Font->Data;		// punta al data array
  uint8_t	*OutData;
  uint8_t	A;
  uint32_t	Len=FontNumOfChars(Font);	// # di char nel font
  
  
  
  if(Font == NULL)				// se non c'e'
    return(NULL);				// propaga NULL
  if(Font->BytePerChar != 8)			// se non ha 8 byte per char
    return(Font);				// non possiamo fare niente
  
  // crea una nuova copia del font per poi distruggere l'originale
  if((Copy=NewFont(Font)) == NULL)		// se non riesce a copiare
  {
    fprintf(stderr,"%s: unable to copy original font.\n",__FUNCTION__);
    free(Font);					// distrugge comunque l'origin.
    return(NULL);				// termian con fallimento
  }
  
  // scambia width ed height
  A=		Copy->Width;
  Copy->Width=	Copy->Height;
  Copy->Height=	A;
  
  // modifica tutti i char uno alla volta
  OutData=Copy->Data;				// punta alla destinazione
  do
  {
    Ccw(InData,OutData);			// ruota a sinistra
    InData+=Font->BytePerChar;			// prossimo char
    OutData+=Font->BytePerChar;
  }while(--Len);				// fino a fine array
  free(Font);					// distrugge l'originle
  return(Copy);					// termina
}



FontDesc_t *CharCut(FontDesc_t *Font,uint32_t Retain)
{
  /* Limits the number of bytes per char. */

  
  
  FontDesc_t	*New;
  uint8_t	*InData=Font->Data;		// punta al data array
  uint8_t	*Data;
  uint32_t	Chars,NewLen,i;
  


  if(Font == NULL)				// se non c'e'
    return(NULL);				// propaga NULL
  if(Retain >= Font->BytePerChar)		// se non taglia
    return(Font);				// lascia tutto com'e'
  
  // alloca un nuovo font con dimensione modificata
  Chars=FontNumOfChars(Font);			// # di char nel font
  printf("%s: Chars: %lu\n",__FUNCTION__,Chars);
  NewLen=Chars * Retain;			// nuova dim.
  if((New=malloc(sizeof(FontDesc_t) + NewLen)) == NULL)
    return(New);				// termina con fallimento
  
  // copia il descrittore originale in RAM
  memcpy(New,Font,sizeof(FontDesc_t));		// copia il descriptor
  Data=(uint8_t*)&New[1];			// punta oltre il descriptor
  New->Data=		Data;			// modifica il puntatore
  New->BytePerChar=	Retain;			// modifica la dim.
//  New->DataLen=		NewLen;

  // copia il data array tagliando ciscun char a Retain byte
  do
  {
    i=Retain;					// carica il contatore
    do
    {
      *Data++=	*InData++;			// copia
    }while(--i);
    InData+=Font->BytePerChar - Retain;		// salta la parte tagliata
  }while(--Chars);
  free(Font);					// distrugge il vecchio
  return(New);					// termina
}



FontDesc_t *CharModify(FontDesc_t *Font,uint8_t Ascii,const uint8_t *Data,
			uint8_t DataLen)
{
  /* Modify font shapes. */



  uint8_t	*Shape;



  if(Font == NULL)				// se non c'e'
    return(Font);				// propaga NULL
  if((Ascii < Font->First) || (Ascii > Font->Last))
    return(Font);				// fermati immediatamente
  Shape=&Font->Data[(Ascii - Font->First) * Font->BytePerChar];
  memcpy(Shape,Data,DataLen);
  return(Font);					// termina
}



FontDesc_t *Collapse(FontDesc_t *Font,uint32_t CollPt)
{
  /* Collapses font shape into a condensed one. */
  
  
  
  uint8_t	*Data=Font->Data;		// punta al data array
  uint32_t	Len=FontDataSize(Font);		// dimensione data array
  
  
  
  if(Font == NULL)				// se non c'e'
    return(NULL);				// propaga NULL
  do
  {
    *Data=ByteCollapse(*Data,CollPt);		// shape collapse
    Data++;					// avanti Savoia!!
  }while(--Len);				// fino a fine array
  return(Font);					// termina


}



FontDesc_t *DuplicateRow(FontDesc_t *Font,uint32_t Row)
{
  /* Generates higer font duplicating one row. */



  FontDesc_t	*New;
  uint8_t	*InData=Font->Data;		// punta al data array
  uint8_t	*Data;
  uint32_t	Chars,NewLen,i;
  


  if(Font == NULL)				// se non c'e'
    return(NULL);				// propaga NULL
  if(Row >= Font->BytePerChar)			// se esce dalla shape
    return(Font);				// lascia tutto com'e'
  
  // alloca un nuovo font con dimensione modificata
  Chars=FontNumOfChars(Font);			// # di char nel font
  printf("%s: Chars: %lu\n",__FUNCTION__,Chars);
  NewLen=Chars * (Font->Height + 1);		// nuova dim.
  if((New=malloc(sizeof(FontDesc_t) + NewLen)) == NULL)
    return(New);				// termina con fallimento
  
  // copia il descrittore originale in RAM
  memcpy(New,Font,sizeof(FontDesc_t));		// copia il descriptor
  Data=(uint8_t*)&New[1];			// punta oltre il descriptor
  New->Data=		Data;			// modifica il puntatore
  New->Height=		Font->Height + 1;	// modifica l'altezza
  New->BytePerChar=	Font->BytePerChar + 1;	// modifica la dim.

  // copia il data array tagliando ciscun char a Retain byte
  do
  {
    for(i=0;i<Font->Height;i++)
    {
      *Data++=	*InData;			// copia
      if(i == Row)				// se e' la riga incriminata
	*Data++=*InData;			// raddoppia la riga
      InData++;
    }
  }while(--Chars);
  free(Font);					// distrugge il vecchio
  return(New);					// termina
}



FontDesc_t *HeightIncrement(FontDesc_t *Font,uint32_t Rows)
{
  /* Adds N blank rows on top. */



  FontDesc_t	*New;
  uint8_t	*InData=Font->Data;		// punta al data array
  uint8_t	*Data;
  uint32_t	Chars,NewLen,i;
  


  if(Font == NULL)				// se non c'e'
    return(NULL);				// propaga NULL
  if(!Rows)					// se non aggiunge niente
    return(Font);				// lascia tutto com'e'
  
  // alloca un nuovo font con dimensione modificata
  Chars=FontNumOfChars(Font);			// # di char nel font
  printf("%s: Chars: %lu\n",__FUNCTION__,Chars);
  NewLen=Chars * (Font->Height + Rows);		// nuova dim.
  if((New=malloc(sizeof(FontDesc_t) + NewLen)) == NULL)
    return(New);				// termina con fallimento
  
  // copia il descrittore originale in RAM
  memcpy(New,Font,sizeof(FontDesc_t));		// copia il descriptor
  Data=(uint8_t*)&New[1];			// punta oltre il descriptor
  New->Data=		Data;			// modifica il puntatore
  New->Height=		Font->Height + Rows;	// modifica l'altezza
  New->BytePerChar=	Font->BytePerChar + Rows;	// modifica la dim.

  // copia il data array tagliando ciscun char a Retain byte
  do
  {
    for(i=0;i<Rows;i++)
      *Data++=0;				// aggiunge righe vuote
    for(i=0;i<Font->Height;i++)
      *Data++=	*InData++;			// copia l'originale
  }while(--Chars);
  free(Font);					// distrugge il vecchio
  return(New);					// termina
}



FontDesc_t *FirstModify(FontDesc_t *Font,uint32_t First)
{
  /* Changes first ASCII code available inside the font. */



  FontDesc_t	*New;
  uint8_t	*InData=Font->Data;		// punta al data array
  uint8_t	*Data;
  uint32_t	Chars,NewLen,i,NewChars,DelChars;
  


  if(Font == NULL)				// se non c'e'
    return(NULL);				// propaga NULL
  if(First > Font->Last)			// evita l'inversione
    First=Font->Last;
  
  // alloca un nuovo font con dimensione modificata
  Chars=FontNumOfChars(Font);			// # di char nel font
  NewLen=(Font->Last - First + 1) * Font->BytePerChar;	// nuova dim.
  if((New=malloc(sizeof(FontDesc_t) + NewLen)) == NULL)
    return(New);				// termina con fallimento
  
  // copia il descrittore originale in RAM
  memcpy(New,Font,sizeof(FontDesc_t));		// copia il descriptor
  Data=(uint8_t*)&New[1];			// punta oltre il descriptor
  New->Data=		Data;			// modifica il puntatore
  New->First=		First;			// modifica il primo ASCII

  // se First < Font->First allora allunga il font set
  if(First < Font->First)
  {
    // inserisce il pezzo mancante se First < Font->First
    NewChars=Font->First - First;		// # nuovi char
    do
    {
      for(i=0;i<Font->BytePerChar;i++)
	*Data++=0;				// aggiunge righe vuote
    }while(NewChars--);

    // copia il data array originale
    do
    {
      
      for(i=0;i<Font->BytePerChar;i++)
	*Data++=	*InData++;		// copia l'originale
    }while(--Chars);
  }
  else
  {
    // accorcia il font set originale
    DelChars=First - Font->First;		// # char cancellati
    Chars-=DelChars;
    InData=&InData[DelChars * Font->BytePerChar];	// salta
    do
    {
      for(i=0;i<Font->BytePerChar;i++)
	*Data++=	*InData++;		// copia l'originale
    }while(--Chars);
  }
  free(Font);					// distrugge il vecchio
  return(New);					// termina
}



FontDesc_t *LastModify(FontDesc_t *Font,uint32_t Last)
{
  /* Changes last ASCII code available inside the font. */



  FontDesc_t	*New;
  uint8_t	*InData=Font->Data;		// punta al data array
  uint8_t	*Data;
  uint32_t	Chars,NewLen,i,NewChars,DelChars;
  


  if(Font == NULL)				// se non c'e'
    return(NULL);				// propaga NULL
  if(Last < Font->First)			// evita l'inversione
    Last=Font->First;
  if(Last > 255)				// ci fermiamo a 256 char
    Last=255;
  
  // alloca un nuovo font con dimensione modificata
  Chars=FontNumOfChars(Font);			// # di char nel font
  NewLen=(Last - Font->First + 1) * Font->BytePerChar;	// nuova dim.
  if((New=malloc(sizeof(FontDesc_t) + NewLen)) == NULL)
    return(New);				// termina con fallimento
  
  // copia il descrittore originale in RAM
  memcpy(New,Font,sizeof(FontDesc_t));		// copia il descriptor
  Data=(uint8_t*)&New[1];			// punta oltre il descriptor
  New->Data=		Data;			// modifica il puntatore
  New->Last=		Last;			// modifica l'ultimo ASCII

  // se Last > Font->Last allora allunga il font set
  if(Last > Font->Last)
  {
    // copia il data array originale
    do
    {
      for(i=0;i<Font->BytePerChar;i++)
	*Data++=	*InData++;		// copia l'originale
    }while(--Chars);
    
    // inserisce il pezzo mancante se Last > Font->Last
    if(Last > Font->Last)
    {
      NewChars=Last - Font->Last;		// # nuovi char
      do
      {
	for(i=0;i<Font->BytePerChar;i++)
	  *Data++=0;				// aggiunge righe vuote
      }while(NewChars--);
    }
  }
  else
  {
    // accorcia il data set originale
    DelChars=Font->Last - Last;			// # char cancellati
    Chars-=DelChars;
    do
    {
      for(i=0;i<Font->BytePerChar;i++)
	*Data++=	*InData++;		// copia l'originale
    }while(--Chars);
    
  }
  free(Font);					// distrugge il vecchio
  return(New);					// termina
}

/* $Source$ */


