/*
    Font filters implementation.
    Copyright (C) 2020 2021  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */

/**  \file filters.h
     \author Alessandro Accattini
     \date 2020
     \brief Font filters implementation.
 */



#ifndef __FILTERS_H
#define __FILTERS_H		1


#include "fontdesc.h"				// font descriptor data type
   
/***************************************************************************/
/* define pubblici */

/***************************************************************************/
/* macro pubbliche */

/***************************************************************************/
/* strutture pubbliche */

/***************************************************************************/
/* costanti pubbliche */

/***************************************************************************/
/* variabili pubbliche */

/***************************************************************************/
/* funzioni pubbliche */

FontDesc_t *NewFont(const FontDesc_t *Font);
  /**< \brief Allocates a new modifiable font.
       \param Font[in]   Starting font.
       \return           An in-place modifiable font.

       Allocates a new modifiable font. All fonts are declared as constants,
       this function dulicates font declaration in RAM to be in-place
       modifiable. All fonts created by this function can be deallocated
       calling free().
   */


FontDesc_t *RightShift(FontDesc_t *Font,uint32_t Shift);
  /**< \brief Right shift all font data.
       \param Font[in]   In-place modifiable font (see \ref NewFont()).
       \param Shift[in]  Shift amount (in pixel).
       \return           Modified font.
   */

FontDesc_t *LeftShift(FontDesc_t *Font,uint32_t Shift);
  /**< \brief Left shift all font data.
       \param Font[in]   In-place modifiable font (see \ref NewFont()).
       \param Shift[in]  Shift amount (in pixel).
       \return           Modified font.
   */

FontDesc_t *FlipLeftToRight(FontDesc_t *Font);
  /**< \brief Flips characters left to right.
       \param Font[in]   In-place modifiable font (see \ref NewFont()).
       \return           Modified font.
   */

FontDesc_t *RotateRight(FontDesc_t *Font);
  /**< \brief Rotates all characters shapes 90 degrees clock-wise.
       \param Font[in]   In-place modifiable font (see \ref NewFont()).
       \return           Modified font.
   */

FontDesc_t *RotateLeft(FontDesc_t *Font);
  /**< \brief Rotates all characters shapes 90 degrees counter-clock-wise.
       \param Font[in]   In-place modifiable font (see \ref NewFont()).
       \return           Modified font.
   */

FontDesc_t *CharCut(FontDesc_t *Font,uint32_t Retain);
  /**< \brief Limits the number of bytes per char.
       \param Font[in]   In-place modifiable font (see \ref NewFont()).
       \param Retain[in] Number of bytes to retain for each char.
       \return           Modified font.
   */

FontDesc_t *CharModify(FontDesc_t *Font,uint8_t Ascii,const uint8_t *Data,
			uint8_t DataLen);
  /**< \brief Modify font shapes.
       \param Font[in]     In-place modifiable font (see \ref NewFont()).
       \param Ascii[in]    First ASCII code to modify.
       \param Data[in]     New data for given ASCII code.
       \param DataLen[in]  Numbr of bytes in /p Data.
       \return             Modified font.
   */

FontDesc_t *Collapse(FontDesc_t *Font,uint32_t CollPt);
  /**< \brief Collapses font shape into a condensed one.
       \param Font[in]   In-place modifiable font (see \ref NewFont()).
       \param CollPt[in] Collapse point.
       \return           Modified font.

       Collapses font shape into a condensed one.
       All pixels on column \p CollPt are deleted and font shape is collapsed
       in a condensed one.
   */

FontDesc_t *DuplicateRow(FontDesc_t *Font,uint32_t Row);
  /**< \brief Generates higer font duplicating one row.
       \param Font[in]   In-place modifiable font (see \ref NewFont()).
       \param Row[in]    Font row to be doubled.
       \return           Modified font.
   */

FontDesc_t *HeightIncrement(FontDesc_t *Font,uint32_t Rows);
  /**< \brief Adds N blank rows on top.
       \param Font[in]   In-place modifiable font (see \ref NewFont()).
       \param Rows[in]   Number of new rows to add.
       \return           Modified font.
   */

FontDesc_t *FirstModify(FontDesc_t *Font,uint32_t First);
  /**< \brief Changes first ASCII code available inside the font.
       \param Font[in]   In-place modifiable font (see \ref NewFont()).
       \param First[in]  First available ASCII code.
       \return           Modified font.
   */

FontDesc_t *LastModify(FontDesc_t *Font,uint32_t Last);
  /**< \brief Changes last ASCII code available inside the font.
       \param Font[in]   In-place modifiable font (see \ref NewFont()).
       \param Last[in]   Last available ASCII code.
       \return           Modified font.
   */

/***************************************************************************/
#endif

/* $Source$ */

