/*
    Font printing functions.
    Copyright (C) 2020 2021  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */


/***************************************************************************/
/* include vari */

#include <stdio.h>			// printf()
#include "fontdesc.h"			// font data type

/***************************************************************************/
/* define interni alla libreria */

/***************************************************************************/
/* strutture contenute all'interno della libreria */

typedef struct
{
  uint8_t	OffsetMsb;		// parte alta dell'offset
  uint8_t	Width;			// char width
  uint16_t	OffsetLsb;		// parte bassa dell'offset
}MicroCharEntry_t;

/***************************************************************************/
/* costanti interne alla libreria */

/***************************************************************************/
/* costanti pubbliche */
			
/***************************************************************************/
/* variabili interne alla libreria */

/***************************************************************************/
/* funzioni interne alla libreria */

static void MicrochipTable(const FontDesc_t *Font);
  /* Stampa la tabella dei char richiesta dal formato Microchip. */

static char *LegacyDataShape(const uint8_t *Data,uint32_t Len);
  /* Ritorna la stringa che rappresenta Len byte prelevati da Data. */

static char *DataShape(const uint8_t *Data,uint32_t Len);
  /* Ritorna la stringa che rappresenta Len byte prelevati da Data. */

static char *UndefinedShape(uint8_t Width);
  /* Ritorna una shape 'undefined' di larghezza indicata. */

/***************************************************************************/
/* funzioni */

static void MicrochipTable(const FontDesc_t *Font)
{
  /* Stampa la tabella dei char richiesta dal formato Microchip. */
  
  /*
     Di fatto la tabella punta a delle bitmap che hanno tutte le stesse
     dimensioni (in quanto ricavata da un font monospace).
  */
  
  
  uint32_t	Offset;
  uint32_t	Entries=FontNumOfChars(Font);
  char		Ascii;
  
  
  
  // determina l'offset iniziale
  Offset=Entries * 4;				// lunghezza tabella
  Offset+=8;					// lunghezza header Microchip
  
  // stampa la tabella dei char
  Ascii=Font->First;
  do
  {
    printf("\t\t\t0x%02hhx,0x%02hhx,0x%02hhx,0x%02hhx,\t// '%c'\n",
	   (uint8_t)(Offset>>16),		// offset Msb
	   Font->Width,
	   (uint8_t)Offset,			// offset Lsb L
	   (uint8_t)(Offset>>8),		// offset Lsb H
	   ((Ascii >= ' ') && (Ascii <= 0x7e)) ? Ascii : ' ');
    Ascii++;					// avanti Savoia!!
    Offset+=(uint32_t)Font->BytePerChar;
  }while(--Entries);
  putchar('\n');
  return;					// termina
}



static char *LegacyDataShape(const uint8_t *Data,uint32_t Len)
{
  /* Ritorna la stringa che rappresenta Len byte prelevati da Data. */



  static char		Shape[32+2];
  char			*Ptr=Shape;
  unsigned int		i;



  if(Len > 4)					// se e' piu' di 4 byte
    Len=4;					// fermati a 4 byte
  do
  {
    for(i=0;i<8;i++)
    {
      *Ptr++=(*Data & (0x80>>i)) ? '#' : '.';
    }
    Data++;					// next data byte
  }while(--Len);				// per tutti i byte passati
  *Ptr=0x00;					// termina la stringa
  return(Shape);				// termina
}



static char *DataShape(const uint8_t *Data,uint32_t Len)
{
  /* Ritorna la stringa che rappresenta Len byte prelevati da Data. */



  static char		Shape[32+2];
  char			*Ptr=Shape;
  unsigned int		i;



  if(Len > 4)					// se e' piu' di 4 byte
    Len=4;					// fermati a 4 byte
  do
  {
    for(i=0;i<8;i++)
    {
      *Ptr++=(*Data & (0x01<<i)) ? '#' : '.';
    }
    Data++;					// next data byte
  }while(--Len);				// per tutti i byte passati
  *Ptr=0x00;					// termina la stringa
  return(Shape);				// termina
}



static char *UndefinedShape(uint8_t Width)
{
  /* Ritorna una shape 'undefined' di larghezza indicata. */
  
  
  
  static char		Shape[32+2];
  char			*Ptr=Shape;



  if(Width > 32)
    Width=32;					// contiene i disastri
  while(Width--)
    *Ptr++='x';					// riempie la stringa
  *Ptr=0x00;					// termina la stringa
  return(Shape);				// termina
}





void PrintCcode(const FontDesc_t *Font,const char *FontName)
{
  /* Prints font as C source code. */
  

  
  unsigned char		*Data;
  unsigned int		i;
  unsigned char		Ascii;

  
  
  // verifica che il font esista
  if(Font == NULL)
  {
    puts("<NULL>");
    return;					// fermati qui
  }

  // stampa il font header
  printf("unsigned char\t%s[:>={\n",FontName);
  printf("\t\t\t0x%02hhx,\t// first ASCII code\n",Font->First);
  printf("\t\t\t0x%02hhx,\t// last ASCII code\n",Font->Last);
  printf("\t\t\t0x%02hhx,\t// char width (in pixel)\n",Font->Width);
  printf("\t\t\t0x%02hhx,\t// char height (in pixel)\n\n\n",Font->Height);

  // stampa tutti i char nel font
  Ascii=Font->First;				// primo codice ASCII
  Data=Font->Data;				// punta al data array
  do
  {
    // stampa l'ASCII code
    printf("\t\t// 0x%02hhx\n",Ascii++);	// ASCII code

    // stampa la char shape
    i=Font->BytePerChar;			// carica il byte counter
    do
    {
      printf("\t\t\t0x%02hhx,\t//  %s\n",*Data,DataShape(Data,1));
      Data++;
    }while(--i);
    putchar('\n');				// lascia una riga vuota
  }while(Ascii && (Ascii <= Font->Last));
}



void PrintClegacy(const FontDesc_t *Font,const char *FontName)
{
  /* Prints font as C source code in legacy endianness format. */



  unsigned char		*Data;
  unsigned int		i;
  unsigned char		Ascii;

  
  
  // verifica che il font esista
  if(Font == NULL)
  {
    puts("<NULL>");
    return;					// fermati qui
  }

  // stampa il font header
  printf("unsigned char\t%s[:>={\n",FontName);
  printf("\t\t\t0x%02hhx,\t// first ASCII code\n",Font->First);
  printf("\t\t\t0x%02hhx,\t// last ASCII code\n",Font->Last);
  printf("\t\t\t0x%02hhx,\t// char width (in pixel)\n",Font->Width);
  printf("\t\t\t0x%02hhx,\t// char height (in pixel)\n\n\n",Font->Height);

  // stampa tutti i char nel font
  Ascii=Font->First;				// primo codice ASCII
  Data=Font->Data;				// punta al data array
  do
  {
    // stampa l'ASCII code
    printf("\t\t// 0x%02hhx\n",Ascii++);	// ASCII code

    // stampa la char shape
    i=Font->BytePerChar;			// carica il byte counter
    do
    {
      printf("\t\t\t0x%02hhx,\t//  %s\n",*Data,LegacyDataShape(Data,1));
      Data++;
    }while(--i);
    putchar('\n');				// lascia una riga vuota
  }while(Ascii && (Ascii <= Font->Last));
}



void PrintTcl(const FontDesc_t *Font,const char *FontName)
{
  /* Prints font as TCL code. */



  unsigned char		*Data1,*Data2;
  unsigned int		i;
  unsigned char		Ascii;



  // verifica che il font esista
  if(Font == NULL)
  {
    puts("<NULL>");
    return;					// fermati qui
  }

  // stampa il font header
  printf("set %s \"\"\n",FontName);
  printf("append %s\t\\x%02hhx\\x%02hhx\\x%02hhx\\x%02hhx\n\n",
  		FontName,			// font variable name
  		Font->First,			// first ASCII code
		Font->Last,			// last ASCII code
		Font->Width,			// char width
		Font->Height);			// char height

  // stampa tutti i char nel font
  Ascii=Font->First;				// primo codice ASCII
  Data1=Data2=Font->Data;			// punta al data array
  do
  {
    // stampa l'ASCII code
    printf("\t\t;# 0x%02hhx\n",Ascii++);	// ASCII code

    // stampa la char shape
    i=Font->BytePerChar;			// carica il byte counter
    do
    {
      printf("\t\t\t;#  %s\n",DataShape(Data1++,1));
    }while(--i);

    // stampa l'append line
    printf("append %s ",FontName);
    i=Font->BytePerChar;			// carica il byte counter
    do
    {
      printf("\\x%02hhx",*Data2++);
    }while(--i);
    putchar('\n');				// fine riga
    putchar('\n');				// lascia una riga vuota
  }while(Ascii && (Ascii <= Font->Last));
}



void PrintAvra(const FontDesc_t *Font,const char *FontName)
{
  /* Prints font as Avra ASM source code. */
  
  
  
  // verifica che il font esista
  if(Font == NULL)
  {
    puts("<NULL>");
    return;					// fermati qui
  }

  unsigned char		*Data;
  unsigned int		i;
  unsigned char		Ascii;



  // verifica che il font esista
  if(Font == NULL)
  {
    puts("<NULL>");
    return;					// fermati qui
  }

  // stampa il font header
  printf("%s:\n",FontName);
  printf("\t\t.db 0x%02hhx,0x%02hhx\t; first ASCII, last ASCII\n",
	 Font->First,Font->Last);
  printf("\t\t.db 0x%02hhx,0x%02hhx\t; char width, char height\n\n",
	 Font->Width,Font->Height);

  // stampa tutti i char nel font
  Ascii=Font->First;				// primo codice ASCII
  Data=Font->Data;				// punta al data array
  do
  {
    // stampa l'ASCII code
    printf("\t; 0x%02hhx\n",Ascii++);		// ASCII code

    // stampa la char shape
    i=Font->BytePerChar / 2;			// carica il row counter
    do
    {
      printf("\t\t.db 0x%02hhx,0x%02hhx\t;  %s\n",
	     Data[0],Data[1],DataShape(Data,1));
      Data++;
      printf("\t\t\t\t;  %s\n",DataShape(Data,1));
      Data++;
    }while(--i);
    putchar('\n');				// lascia una riga vuota
  }while(Ascii && (Ascii <= Font->Last));
}



void PrintMicrochip(const FontDesc_t *Font,const char *FontName)
{
  /* Prints font in Microchip code. */



  unsigned char		*Data;
  unsigned int		i;
  unsigned char		Ascii;

  
  
  // verifica che il font esista
  if(Font == NULL)
  {
    puts("<NULL>");
    return;					// fermati qui
  }

  // stampa il font header
  printf("unsigned char\t%s[:>={\n",FontName);
  printf("\t\t\t0x%02hhx,0x%02hhx,\t// reserved, font ID\n",(uint8_t)0,(uint8_t)0);
  printf("\t\t\t0x%02hhx,0x%02hhx,\t// first ASCII code\n",Font->First,(uint8_t)0);
  printf("\t\t\t0x%02hhx,0x%02hhx,\t// last ASCII code\n",Font->Last,(uint8_t)0);
  printf("\t\t\t0x%02hhx,0x%02hhx,\t// reserved, height\n\n",(uint8_t)0,Font->Height);
  
  // stampa la tabella dei char
  MicrochipTable(Font);

  // stampa tutti i char nel font
  Ascii=Font->First;				// primo codice ASCII
  Data=Font->Data;				// punta al data array
  do
  {
    // stampa l'ASCII code
    printf("\t\t// 0x%02hhx\n",Ascii++);	// ASCII code

    // stampa la char shape
    i=Font->BytePerChar;			// carica il byte counter
    do
    {
      printf("\t\t\t0x%02hhx,\t//  %s\n",*Data,LegacyDataShape(Data,1));
      Data++;
    }while(--i);
    putchar('\n');				// lascia una riga vuota
  }while(Ascii && (Ascii <= Font->Last));
}



void PrintText(const FontDesc_t *Font,const char *Text,uint8_t Len)
{
  /* Prints on stdout using given font. */



  const	uint8_t	*Data;				// shape data ptr
  const	char	*Shape;				// shape to print
  uint8_t	i,j;				// current line



  for(i=0;i<Font->Height;i++)
  {
    for(j=0;j<Len;j++)
    {
      if((Text[j] < Font->First) || (Text[j] > Font->Last))
      {
        Shape=UndefinedShape(Font->Width);
      }
      else
      {
	Data=&Font->Data[(Text[j] - Font->First) * Font->BytePerChar + i];
        Shape=DataShape(Data,1);		// shape to print
      }
      printf("%s",Shape);			// shape on stdout
    }
    putchar('\n');				// end of line
  }
  return;					// termina
}

/* $Source$ */




